/*
 * DNSResponder - poses as a DNS server and returns useful stuff
 *
 * Copyright 2007 (c) Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/queue.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <err.h>
#include <fcntl.h>
#include <netdb.h>
#include <unistd.h>
#include <assert.h>

#include <event.h>
#include <evdns.h>
#include <dnet.h>

#include "dnsdata.gen.h"
#include "utils.h"

#define DNSRESPONDER_TAG	0xae

static int result_fd = -1;
static char *query_suffix = ".osd.honeyd.org";

static struct addr local_addr;

static void
dns_callback(struct evdns_server_request *req, void *arg)
{
	struct evdns_server_question *q;
	struct sockaddr sa;
	struct addr addr;
	int i;

	evdns_server_request_get_requesting_addr(req, &sa, sizeof(sa));
	addr_ston(&sa, &addr);
	fprintf(stderr, "Got request from %s (%d qs)\n",
	    addr_ntoa(&addr), req->nquestions);

	for (i = 0; i < req->nquestions; ++i) {
		q = req->questions[i];
		fprintf(stderr, "\t%d %d %s\n", q->type, q->class, q->name);
	}

	if (req->nquestions == 0 || req->nquestions > 2)
		goto error;

	for (i = 0; i < req->nquestions; ++i) {
		q = req->questions[i];

		if (q->type == EVDNS_TYPE_A && q->class == EVDNS_CLASS_INET)
			break;
	}

	if (i >= req->nquestions)
		goto error;

	if (strlen(q->name) < strlen(query_suffix) ||
	    strcmp(q->name + strlen(q->name) - strlen(query_suffix),
		query_suffix) != 0)
		goto error;

	/* log the query */
	if (result_fd != -1) {
		struct evbuffer *buf = evbuffer_new();
		struct DNSResponder *dr = DNSResponder_new();
		struct timeval tv;

		/* logging */
		gettimeofday(&tv, NULL);
		EVTAG_ASSIGN(dr, tv_sec, tv.tv_sec);
		EVTAG_ASSIGN(dr, tv_usec, tv.tv_usec);

		EVTAG_ASSIGN(dr, remote_addr, (uint8_t *)&addr, sizeof(addr));
		EVTAG_ASSIGN(dr, query, q->name);

		evtag_marshal_DNSResponder(buf, DNSRESPONDER_TAG, dr);

		DNSResponder_free(dr);

		while (EVBUFFER_LENGTH(buf) > 0) {
			if (evbuffer_write(buf, result_fd) == -1) {
				if (errno != EAGAIN)
					err(1, "writing buffer");
			}
		}

		evbuffer_free(buf);

		fsync(result_fd);
	}

	if (strncmp(q->name, "ns.", 3) == 0) {
		evdns_server_request_add_a_reply(req, q->name,
		    1, &local_addr.addr_ip, 60);
		evdns_server_request_add_reply(req,
		    EVDNS_AUTHORITY_SECTION, query_suffix + 1,
		    EVDNS_TYPE_NS, EVDNS_CLASS_INET, 28800, -1, 1,
		    q->name);
	} else {
		evdns_server_request_add_a_reply(req, q->name,
		    1, &addr.addr_ip, 60);
	}
	evdns_server_request_respond(req, 0);

	return;
error:
	evdns_server_request_respond(req, DNS_ERR_SERVERFAILED);
}

static struct evdns_server_port *
add_dns_port(char *argument)
{
	char *ip, *port;
	int fd;

	ip = strsep(&argument, ":");
	if (ip == NULL)
		goto error;

	if (addr_aton(ip, &local_addr) == -1)
		goto error;

	port = strsep(&argument, ":");
	if (port == NULL)
		goto error;

	if (argument != NULL && *argument != '\0')
		goto error;

	fd = make_socket(1, ip, atoi(port));
	if (fd == -1)
		err(1, "make_socket: %s:%s", ip, port);

	return (evdns_add_server_port(fd, 0/* is_tcp */, dns_callback, NULL));

error:
	errx(1, "Illegal ipaddr:port specified");
	/* NOTREACHED */
}

static void
usage(const char *name)
{
	fprintf(stderr, 
	    "%s: [-e excludefile] [-i if] <IP address|network>...\n\n"
	    "\t-V          print version number,\n"
	    "\t-i <alias>  specifies the IP alias to connect from,\n",
	    name);
}

int
main(int argc, char **argv)
{
	const char *name = argv[0];
	const char *output = NULL;
	int ch;

	event_init();
	evtag_init();

	while ((ch = getopt(argc, argv, "Q:Vo:i:")) != -1)
		switch(ch) {
		case 'Q':
			if (optarg[0] != '.')
				errx(1, "Bad query suffix: %s", optarg);
			query_suffix = optarg;
			break;
		case 'o':
			output = optarg;
			break;
		case 'i':
			add_dns_port(optarg);
			break;
		case 'V':
			fprintf(stderr, "DNSResponder %s\n", VERSION);
			exit(0);
		case 'h':
		default:
			usage(name);
			exit(1);
		}

	if (output) {
		fprintf(stderr, "... Writing results to \"%s\"\n", output);
		result_fd = open(output, O_WRONLY | O_CREAT | O_APPEND, 0600);
		if (result_fd == -1)
			err(1, "cannot open \"%s\"", output);
	}

	argc -= optind;
	argv += optind;

	if (argc) {
		fprintf(stderr, "Superflous arguments\n");
		exit(0);
	}

	event_dispatch();

	return (0);
}
