/*
 * Copyright 2007 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/queue.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <assert.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <event.h>
#include <evhttp.h>
#include <dnet.h>

#include "dnsscan.h"
#include "dnsdata.gen.h"
#include "state.h"
#include "dnscore.h"
#include "utils.h"
#include "fpdns.h"
#include "fpdns_states.h"
#include "evutils.h"
#include "web.h"

extern const char *query_suffix;

extern struct requestq http_requests;
extern int num_http_requests;
extern int num_pending_http_requests;

extern int use_http;

extern struct history_buffer *history;
extern struct event_histogram *hist_iquery;
extern struct event_histogram *hist_tc;

extern struct event_counter *ctr_completed;
extern struct event_counter *ctr_resolver;
extern struct event_counter *ctr_recursive;
extern struct event_counter *ctr_chaos_answered;
extern struct event_counter *ctr_chaos_empty;
extern struct event_counter *ctr_web_server;

static void request_http_cb(struct evhttp_request *http_req, void *arg);

/*
 * all the fpdns state magic code
 */

/* states that a request can be in */
enum FPDNS_STATES {
	OPEN_RESOLVER = 0,
	CHAOS_PROBE = 1,
	FPDNS_IQUERY = 2,
	FPDNS_TC = 3,
	NUM_STATES = 4
};

static int
probe_init_open_resolver(struct state *state, struct request *req,
    uint8_t *buf, size_t buf_len)
{
	static char ascii[512];
	static char name[512];

	bin_to_ascii((uint8_t *)&req->dest.addr_ip, sizeof(req->dest.addr_ip),
	    ascii, sizeof(ascii));
	snprintf(name, sizeof(name), "%s%s", ascii, query_suffix);

	return (dnscore_request_data_build(name, strlen(name),
		req->id, EVDNS_TYPE_A, EVDNS_CLASS_INET, buf, buf_len));
}

static int
next_state_open_resolver(struct state *state, struct request *req,
    struct addr *addr, uint8_t *buf, size_t buf_len)
{
	struct evbuffer *evbuf = NULL;
	struct dns_reply reply;

	if (buf == NULL) {
		/* timed out */
		return (ABORT_STATE);
	}

	/* got data */
	if (dnscore_reply_parse(EVDNS_TYPE_A, &reply, buf, buf_len) != 0)
		goto error;

	req->payload_size = buf_len;
	req->payload = malloc(buf_len);
	assert(req->payload != NULL);
	memcpy(req->payload, buf, buf_len);

	/* analyze the reply */
	if (reply.data.a.addrcount == 1) {
		req->reply.addr_type = ADDR_TYPE_IP;
		req->reply.addr_bits = IP_ADDR_BITS;
		req->reply.addr_ip = reply.data.a.addresses[0];
		req->reply_set = 1;

		evbuf = evbuffer_new();
		assert(evbuf != NULL);
		evbuffer_add_printf(evbuf,
		    "Open resolver %s answered <b>%s</b>\n",
		    addr_ntoa(&req->dest), addr_ntoa(&req->reply));
		history_enter_buffer(history, evbuf);

		event_counter_add(ctr_recursive, 1);
	} else {
		int res = dnscore_reply_parse(EVDNS_TYPE_NS, 
		    &reply, buf, buf_len);
		
		evbuf = evbuffer_new();
		assert(evbuf != NULL);
		evbuffer_add_printf(evbuf, "Open resolver %s answered %s\n",
		    addr_ntoa(&req->dest),
		    res == 0 && reply.have_answer ? 
		    reply.data.ns.name : "nothing");
		history_enter_buffer(history, evbuf);
	}

	event_counter_add(ctr_resolver, 1);

	return (CHAOS_PROBE);

error:
	evbuf = evbuffer_new();
	assert(evbuf != NULL);
	evbuffer_add_printf(evbuf,
	    "Error parsing %ld bytes from %s (expected %s)\n",
	    buf_len, addr_ntoa(addr), addr_ntoa(&req->dest));
	history_enter_buffer(history, evbuf);

	/*
	 * if we cannot parse an answer it's unlikely that we can parse it
	 * next time.  force a timeout.
	 */
	req->retries = DNS_MAX_REQUEST_RETRIES;
	return (state->state);
}

static int
probe_init_chaos(struct state *state, struct request *req,
    uint8_t *buf, size_t buf_len)
{
	const char *name = "version.bind";
	return (dnscore_request_data_build(name, strlen(name),
		req->id, EVDNS_TYPE_TXT, EVDNS_CLASS_CHAOS, buf, buf_len));
}

static int
next_state_chaos(struct state *state, struct request *req,
    struct addr *addr, uint8_t *buf, size_t buf_len)
{
	struct evbuffer *evbuf = NULL;
	struct dns_reply reply;

	if (buf == NULL) {
		/* timed out - we need to move to the next version  */
		goto done;
	}

	/* got data */
	if (dnscore_reply_parse(EVDNS_TYPE_TXT, &reply, buf, buf_len) != 0)
		goto done;

	event_counter_add(ctr_chaos_answered, 1);

	req->chaos_payload_size = buf_len;
	req->chaos_payload = malloc(buf_len);
	assert(req->chaos_payload != NULL);
	memcpy(req->chaos_payload, buf, buf_len);

	req->chaos_reply = strdup(reply.data.text.text);
	assert(req->chaos_reply);

	if (!strlen(req->chaos_reply))
		event_counter_add(ctr_chaos_empty, 1);

	evbuf = evbuffer_new();
	assert(evbuf != NULL);
	evbuffer_add_printf(evbuf,
	    "Resolver %s answered to chaos <b>\"%s\"</b>\n",
	    addr_ntoa(&req->dest),  req->chaos_reply);
	history_enter_buffer(history, evbuf);

done:
	return (FPDNS_IQUERY);
}

static int
probe_init_fpdns_iquery(struct state *state, struct request *req,
    uint8_t *buf, size_t buf_len)
{
	return (fpdns_probe_iquery(buf, buf_len, req->id, query_suffix + 1));
}

static int
next_state_fpdns_iquery(struct state *state, struct request *req,
    struct addr *addr, uint8_t *buf, size_t buf_len)
{
	static char ascii[30];
	struct evbuffer *evbuf = NULL;

	if (fpdns_print(buf, buf_len, ascii, sizeof(ascii)) == -1)
		goto done;

	req->fpdns_iquery = strdup(ascii);
	assert(req->fpdns_iquery != NULL);

	evbuf = evbuffer_new();
	assert(evbuf != NULL);
	evbuffer_add_printf(evbuf,
	    "Resolver %s fpdns iquery: <b>%s</b>\n",
	    addr_ntoa(&req->dest),  ascii);
	history_enter_buffer(history, evbuf);

	event_histogram_add(hist_iquery, ascii, 1);

done:
	return (FPDNS_TC);
}

static int
probe_init_fpdns_tc(struct state *state, struct request *req,
    uint8_t *buf, size_t buf_len)
{
	static char query[128];
	if (!strlen(query))
		snprintf(query, sizeof(query), "tc%s", query_suffix);
	return (fpdns_probe_tc(buf, buf_len, req->id, query));
}

static int
next_state_fpdns_tc(struct state *state, struct request *req,
    struct addr *addr, uint8_t *buf, size_t buf_len)
{
	static char ascii[30];
	struct evbuffer *evbuf = NULL;

	if (fpdns_print(buf, buf_len, ascii, sizeof(ascii)) == -1)
		goto done;

	req->fpdns_tc = strdup(ascii);
	assert(req->fpdns_tc != NULL);

	evbuf = evbuffer_new();
	assert(evbuf != NULL);
	evbuffer_add_printf(evbuf,
	    "Resolver %s fpdns tc: <b>%s</b>\n",
	    addr_ntoa(&req->dest),  ascii);
	history_enter_buffer(history, evbuf);

	event_histogram_add(hist_tc, ascii, 1);

done:
	return (END_STATE);
}

static void
request_launch_get(struct request *req,
    void (*cb)(struct evhttp_request *, void *))
{
	struct evhttp_request *fetch;
	req->ev_con = evhttp_connection_new(addr_ntoa(&req->dest), 80);
	assert(req->ev_con != NULL);
	evhttp_connection_set_local_address(req->ev_con, req->inf->name);
	evhttp_connection_set_retries(req->ev_con, 0);
	evhttp_connection_set_timeout(req->ev_con, 5);

	fetch = evhttp_request_new(cb, req);
	assert(fetch != NULL);

	evhttp_add_header(fetch->output_headers,
	    "Accept",
	    "image/gif, image/x-xbitmap, image/jpeg, "
	    "image/pjeg, application/x-shockwave-flash, /*/*");
	evhttp_add_header(fetch->output_headers,
	    "Accept-Language", "en-us");
	evhttp_add_header(fetch->output_headers,
	    "Accept-Encoding", "gzip, deflate");
	evhttp_add_header(fetch->output_headers,
	    "User-Agent",
	    "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1)");
	evhttp_add_header(fetch->output_headers,
	    "Host", addr_ntoa(&req->dest));

	if (evhttp_make_request(req->ev_con, fetch,
		EVHTTP_REQ_GET, "/") == -1) {
		/* error condition */
		(*cb)(NULL, req);
	}
}

static void
request_http_cb(struct evhttp_request *http_req, void *arg)
{
	struct request *req = arg;

	--num_http_requests;

	if (http_req != NULL &&
	    EVBUFFER_LENGTH(http_req->input_buffer) > 0) {
		struct evbuffer *buf = evbuffer_new();
		char *escaped;
		const char *date =
		    evhttp_find_header(http_req->input_headers, "Date");
		const char *server =
		    evhttp_find_header(http_req->input_headers, "Server");
		const char *powered =
		    evhttp_find_header(http_req->input_headers,
			"X-Powered-By");
		unsigned long len = EVBUFFER_LENGTH(http_req->input_buffer);
		assert(buf != NULL);

		/* we completed a request */
		if (date) {
			DFPRINTF(2, (stderr, "HTTP from %s: Date: %s\n",
				addr_ntoa(&req->dest), date));
			req->fphttp_date = strdup(date);
			assert(req->fphttp_date);
		}
		if (server) {
			DFPRINTF(2, (stderr, "HTTP from %s: Server: %s\n",
				addr_ntoa(&req->dest), server));
			req->fphttp_server = strdup(server);
			assert(req->fphttp_server);
		}
		if (powered) {
			DFPRINTF(2, (stderr,
				"HTTP from %s: X-Powered-By: %s\n",
				addr_ntoa(&req->dest), powered));
			req->fphttp_powered = strdup(powered);
			assert(req->fphttp_powered);
		}

		DFPRINTF(1, (stderr, "HTTP from %s: Body Length: %ld\n",
			addr_ntoa(&req->dest), len));
		req->fphttp_body_len = len;
		if (len > MAX_HTTP_PAYLOAD_SIZE)
			len = MAX_HTTP_PAYLOAD_SIZE;

		req->http_body_payload = malloc(len);
		assert(req->http_body_payload);
		req->http_body_payload_size = len;
		memcpy(req->http_body_payload,
		    EVBUFFER_DATA(http_req->input_buffer), len);

		event_counter_add(ctr_web_server, 1);

		escaped = server != NULL ? evhttp_htmlescape(server) : NULL;
		evbuffer_add_printf(buf, "%s is a web server <b>%s</b> (%ld)",
		    addr_ntoa(&req->dest), escaped,
		    EVBUFFER_LENGTH(http_req->input_buffer));
		if (escaped)
			free(escaped);
		history_enter_buffer(history, buf);
		
		/* log more complete HTTP information */
		req->server_reply = HTTPServerReply_new();
		assert(req->server_reply != NULL);
		web_fill_in_data(MAX_HTTP_PAYLOAD_SIZE,
		    http_req, req->server_reply);
	}

	/* log all the information we collected */
	request_log(req);

	request_free(req);

	/* see if there is another http request we should launch */
	if ((req = TAILQ_FIRST(&http_requests)) != NULL) {
		TAILQ_REMOVE(&http_requests, req, next);
		--num_pending_http_requests;
		++num_http_requests;
		request_launch_get(req, request_http_cb);
	}
}

static void
request_result(struct request *req, int success)
{
	event_del(&req->ev);

	if (!success) {
		event_counter_add(ctr_completed, 1);

		request_free(req);
		return;
	}

	if (!use_http) {
		request_log(req);
		request_free(req);
		return;
	}

	if (num_http_requests >= MAX_HTTP_CONNECTIONS) {
		TAILQ_INSERT_TAIL(&http_requests, req, next);
		++num_pending_http_requests;
		return;
	}

	++num_http_requests;
	request_launch_get(req, request_http_cb);
}

static void
abort_state(struct request *req)
{
	request_result(req, 0 /* success */);
}

static void
finish_state(struct request *req)
{
	request_result(req, 1 /* success */);
}

struct statetable *
setup_fpdns_states(void)
{
	struct statetable *statetable =
	    statetable_new(OPEN_RESOLVER, NUM_STATES);

	statetable_add(statetable, OPEN_RESOLVER,
	    probe_init_open_resolver,
	    next_state_open_resolver);

	statetable_add(statetable, CHAOS_PROBE,
	    probe_init_chaos,
	    next_state_chaos);

	statetable_add(statetable, FPDNS_IQUERY,
	    probe_init_fpdns_iquery,
	    next_state_fpdns_iquery);

	statetable_add(statetable, FPDNS_TC,
	    probe_init_fpdns_tc,
	    next_state_fpdns_tc);

	statetable_set_abort(statetable, abort_state);
	statetable_set_finish(statetable, finish_state);

	return (statetable);
}
