/*
 * Copyright 2007 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>

static char translate[] = {
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
	'a', 'b', 'c', 'd', 'e', 'f'
};

static int transoffset(char a) {
        if (a >= '0' && a <= '9')
                return a - '0';
        if (a >= 'a' && a <= 'f')
                return a + 10 - 'a';
        
        return -1;
}

int
bin_to_ascii(uint8_t *dat, size_t datlen, char *buf, size_t buflen)
{
	int i;
	if (datlen * 2 + 1 > buflen)
		return (-1);

        for (i = 0; i < datlen; ++i) {
		buf[2*i] = translate[dat[i] >> 4];
		buf[2*i + 1] = translate[dat[i] & 0xf];
	}       

	buf[datlen*2] = '\0';

	return (0);
}

int
ascii_to_bin(const char* buf, uint8_t *dat, size_t datlen)
{
	int i;
	if (strlen(buf) / 2 > datlen)
		return (-1);
        
        for (i = 0; i < strlen(buf); i += 2) {
                int left = transoffset(buf[i]);
                int right = transoffset(buf[i + 1]);
                
                if (left == -1 || right == -1)
                        return (-1);
                
                *dat++ = (left << 4) + right;
        }
        
        return (0);
}

int
make_socket_ai(int should_bind, struct addrinfo *ai)
{
        struct linger linger;
        int fd, on = 1, r;
	int serrno, n, nsnd = 0, nrcv = 0;
	unsigned int len;

        /* Create listen socket */
        fd = socket(AF_INET, SOCK_DGRAM, 0);
        if (fd == -1) {
                warn("socket");
                return (-1);
        }

#ifdef SO_SNDBUF
	len = sizeof(n);
	if (getsockopt(fd, SOL_SOCKET, SO_SNDBUF, &n, &len) < 0) {
		warn("getsockopt");
		goto out;
	}

	for (n += 128; n < 1048576; n += 128) {
		if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &n, len) < 0) {
			if (errno == ENOBUFS)
				break;
			goto out;
		}
	}

	nsnd = n;
#endif
#ifdef SO_RCVBUF
	len = sizeof(n);
	if (getsockopt(fd, SOL_SOCKET, SO_RCVBUF, &n, &len) < 0) {
		warn("getsockopt");
		goto out;
	}

	for (n += 128; n < 1048576; n += 128) {
		if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &n, len) < 0) {
			if (errno == ENOBUFS)
				break;
			goto out;
		}
	}

	nrcv = n;

	fprintf(stderr, "... socket buffers (%d/%d)\n", nsnd, nrcv);
#endif

	if (fcntl(fd, F_SETFL, O_NONBLOCK) == -1) {
		warn("fcntl(O_NONBLOCK)");
		goto out;
	}

        if (fcntl(fd, F_SETFD, 1) == -1) {
                warn("fcntl(F_SETFD)");
                goto out;
        }

        setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (void *)&on, sizeof(on));
        setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *)&on, sizeof(on));
        linger.l_onoff = 1;
        linger.l_linger = 5;
        setsockopt(fd, SOL_SOCKET, SO_LINGER, (void *)&linger, sizeof(linger));

	if (should_bind)
		r = bind(fd, ai->ai_addr, ai->ai_addrlen);
	else
		r = connect(fd, ai->ai_addr, ai->ai_addrlen);
	if (r == -1) {
		if (errno != EINPROGRESS) {
			goto out;
		}
	}

	return (fd);

 out:
	serrno = errno;
	close(fd);
	errno = serrno;
	return (-1);
}

int
make_socket(int should_bind, const char *address, u_short port)
{
	int fd;
        struct addrinfo ai, *aitop = NULL;
        char strport[NI_MAXSERV];
        int ai_result;

        memset(&ai, 0, sizeof (ai));
        ai.ai_family = AF_INET;
        ai.ai_socktype = SOCK_DGRAM;
        ai.ai_flags = should_bind ? AI_PASSIVE : 0;
        snprintf(strport, sizeof (strport), "%d", port);
        if ((ai_result = getaddrinfo(address, strport, &ai, &aitop)) != 0) {
                if (ai_result == EAI_SYSTEM)
                        warn("getaddrinfo");
                else
                        warnx("getaddrinfo: %s", gai_strerror(ai_result));
		return (-1);
        }

	fd = make_socket_ai(should_bind, aitop);

	freeaddrinfo(aitop);

	return (fd);
}
